/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "carlfw.h"

#include "../include/compiler.h"

static int
new_fix_entry(struct carlfw *fw, struct carl9170_fw_fix_entry *fix_entry)
{
	struct carl9170_fw_fix_desc_v1 *fix;
	unsigned int len;

	len = sizeof(*fix) + sizeof(*fix_entry);
	fix = malloc(len);
	if (!fix)
		return -ENOMEM;

	carl9170_fw_fill_desc(&fix->head, (uint8_t *) FIX_MAGIC,
			      cpu_to_le16(len),
			      CARL9170_FW_FIX_DESC_MIN_VER,
			      CARL9170_FW_FIX_DESC_CUR_VER);

	memcpy(&fix->data[0], fix_entry, sizeof(*fix_entry));

	return carlfw_desc_add_tail(fw, &fix->head);
}

static struct carl9170_fw_fix_entry *
scan_for_similar_fix(struct carl9170_fw_fix_desc_v1 *fix, __le32 address)
{
	unsigned int i, entries;

	entries = (le16_to_cpu(fix->head.length) - sizeof(*fix)) /
		   sizeof(struct carl9170_fw_fix_entry);

	for (i = 0; i < entries; i++) {
		if (address == fix->data[i].address)
			return &fix->data[i];
	}

	return NULL;
}

static int
add_another_fix_entry(struct carlfw *fw, struct carl9170_fw_fix_desc_v1 *fix,
		 struct carl9170_fw_fix_entry *fix_entry)
{
	unsigned int entry;

	fix = carlfw_desc_mod_len(fw, &fix->head, sizeof(*fix_entry));
	if (IS_ERR_OR_NULL(fix))
		return (int) PTR_ERR(fix);

	entry = (le16_to_cpu(fix->head.length) - sizeof(*fix)) /
		sizeof(*fix_entry) - 1;

	memcpy(&fix->data[entry], fix_entry, sizeof(*fix_entry));
	return 0;
}

static int
del_a_fix_entry(struct carlfw *fw, struct carl9170_fw_fix_desc_v1 *fix,
		struct carl9170_fw_fix_entry *victim)
{
	unsigned long off;
	unsigned int rem_len;

	off = (unsigned long) victim - (unsigned long) fix->data;
	rem_len = le16_to_cpu(fix->head.length) - off;

	if (rem_len) {
		unsigned long cont;
		cont = (unsigned long) victim + sizeof(*victim);
		memmove(victim, (void *)cont, rem_len);
	}

	fix = carlfw_desc_mod_len(fw, &fix->head, -sizeof(*victim));
	if (IS_ERR_OR_NULL(fix))
		return (int) PTR_ERR(fix);

	return 0;
}

static void user_education(void)
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "\teeprom_fix FW-FILE SWITCH [ADDRESS [VALUE [MASK]]]\n");

	fprintf(stderr, "\nDescription:\n");
	fprintf(stderr, "\tThis utility manage a set of overrides which "
			"commands the driver\n\tto load customized EEPROM' "
			"data for all specified addresses.\n");

	fprintf(stderr, "\nParameters:\n");
	fprintf(stderr, "\t'FW-FILE'  = firmware file [basename]\n");
	fprintf(stderr, "\t'SWITCH'   = [=|d|D]\n");
	fprintf(stderr, "\t | '='       => add/set value for address\n");
	fprintf(stderr, "\t | 'D'       => removes all EEPROM overrides\n");
	fprintf(stderr, "\t * 'd'       => removed override for 'address'\n");
	fprintf(stderr, "\n\t'ADDRESS'  = location of the EEPROM override\n");
	fprintf(stderr, "\t\t    (a address map can be found in eeprom.h)\n");
	fprintf(stderr, "\n\t'VALUE'    = replacement value\n");
	fprintf(stderr, "\t'MASK'     = mask for the value placement.\n\n");

	exit(EXIT_FAILURE);
}

int main(int argc, char* args[])
{
	struct carl9170_fw_fix_desc_v1 *fix;
	struct carlfw *fw = NULL;
	unsigned int address, value, mask;
	int err = 0;
	struct carl9170_fw_fix_entry fix_entry, *entry = NULL;

	if (argc < 3 || argc > 6) {
		err = -EINVAL;
		goto out;
	}

	fw = carlfw_load(args[1]);
	if (IS_ERR_OR_NULL(fw)) {
		fprintf(stderr, "Failed to open file \"%s\" (%d).\n",
			args[1], (int) PTR_ERR(fw));
		goto out;
	}

	if (argc > 3) {
		err = sscanf(args[3], "%8x", &address);
		if (err != 1) {
			err = -EINVAL;
			goto out;
		}
		if (address & 3) {
			err = -EINVAL;
			goto out;
		}

		fix_entry.address = cpu_to_le32(address);

		if (argc > 4) {
			err = sscanf(args[4], "%8x", &value);
			if (err != 1) {
				err =- EINVAL;
				goto out;
			}

			if (argc > 5) {
				err = sscanf(args[5], "%8x", &mask);
				if (err != 1) {
					err =- EINVAL;
					goto out;
				}
			} else {
				mask = ~value;
			}

			fix_entry.value = cpu_to_le32(value);
			fix_entry.mask = cpu_to_le32(mask);
		}
	}

	fix = carlfw_find_desc(fw, (uint8_t *)FIX_MAGIC, sizeof(*fix),
			       CARL9170_FW_FIX_DESC_CUR_VER);
	switch(args[2][0]) {
	case '=':
		if (!fix) {
			err = new_fix_entry(fw, &fix_entry);
		} else {
			entry = scan_for_similar_fix(fix, fix_entry.address);
			if (entry) {
				entry->mask = fix_entry.mask;
				entry->value = fix_entry.value;
			} else {
				err = add_another_fix_entry(fw, fix, &fix_entry);
			}
		}
		break;

	case 'd':
		if (fix)
			entry = scan_for_similar_fix(fix, fix_entry.address);

		if (!entry) {
			err = -EINVAL;
			goto out;
		}

		err = del_a_fix_entry(fw, fix, entry);
		break;

	case 'D':
		if (!fix) {
			err = 0;
			goto out;
		}

		carlfw_desc_del(fw, &fix->head);
		break;

	default:
		err = -EINVAL;
		goto out;
	}

	err = carlfw_store(fw);
	if (err) {
		fprintf(stderr, "Failed to apply changes (%d).\n", err);
		goto out;
	}

out:
	carlfw_release(fw);

	if (err == -EINVAL)
		user_education();

	return err ? EXIT_FAILURE : EXIT_SUCCESS;
}
